Optimize dm-verity and fsverity using multibuffer hashing
Subsystems (6)
Eric Biggers
maintainerv7 cover Optimize dm-verity and fsverity using multibuffer hashing
On many modern CPUs, it is possible to compute the SHA-256 hash of two equal-length messages in about the same time as a single message, if all the instructions are interleaved. This is because each SHA-256 (and also most other cryptographic hash functions) is inherently serialized and therefore can't always take advantage of the CPU's full throughput.
An earlier attempt to support multibuffer hashing in Linux was based around the ahash API. That approach had some major issues, as does the alternative ahash-based approach proposed by Herbert (see my response at
https://lore.kernel.org/linux-crypto/[email protected]/).
This patchset instead takes a much simpler approach of just adding a synchronous API for hashing equal-length messages.
This works well for dm-verity and fsverity, which use Merkle trees and therefore hash large numbers of equal-length messages.
This patchset is organized as follows:
- Patch 1-2 add crypto_shash_finup_mb() and tests for it.
- Patch 3-4 implement finup_mb on x86_64 and arm64, using an interleaving factor of 2.
- Patch 5 adds multibuffer hashing support to fsverity.
- Patch 6-7 add multibuffer hashing support to dm-verity.
This patchset increases raw SHA-256 hashing throughput by up to 98%, depending on the CPU (see patches for per-CPU results). The throughput of cold-cache reads from dm-verity and fsverity increases by around 35%.
Changed in v7:
- Rebased onto v6.12-rc1 and dropped patches that were upstreamed.
- Added performance results for more CPUs.
Changed in v6:
- All patches: added Reviewed-by and Acked-by tags
- "crypto: testmgr - add tests for finup_mb": Whitespace fix
- "crypto: testmgr - generate power-of-2 lengths more often":
Fixed undefined behavior
- "fsverity: improve performance by using multibuffer hashing":
Simplified a comment
- "dm-verity: reduce scope of real and wanted digests":
Fixed mention of nonexistent function in commit message
- "dm-verity: improve performance by using multibuffer hashing":
Two small optimizations, and simplified a comment
Changed in v5:
- Reworked the dm-verity patches again. Split the preparation work into separate patches, fixed two bugs, and added some new cleanups.
- Other small cleanups
Changed in v4:
- Reorganized the fsverity and dm-verity code to have a unified code path for single-block vs. multi-block processing. For data blocks they now use only crypto_shash_finup_mb().
Changed in v3:
- Change API from finup2x to finup_mb. It now takes arrays of data buffer and output buffers, avoiding hardcoding 2x in the API.
Changed in v2:
- Rebase onto cryptodev/master
- Add more comments to assembly
- Reorganize some of the assembly slightly
- Fix the claimed throughput improvement on arm64
- Fix incorrect kunmap order in fs/verity/verify.c
- Adjust testmgr generation logic slightly
- Explicitly check for INT_MAX before casting unsigned int to int
- Mention SHA3 based parallel hashes
- Mention AVX512-based approach
Eric Biggers (7):
crypto: shash - add support for finup_mb
crypto: testmgr - add tests for finup_mb
crypto: x86/sha256-ni - add support for finup_mb
crypto: arm64/sha256-ce - add support for finup_mb
fsverity: improve performance by using multibuffer hashing
dm-verity: reduce scope of real and wanted digests
dm-verity: improve performance by using multibuffer hashing
arch/arm64/crypto/sha2-ce-core.S | 281 ++++++++++++++++++++- arch/arm64/crypto/sha2-ce-glue.c | 40 +++ arch/x86/crypto/sha256_ni_asm.S | 368 ++++++++++++++++++++++++++++ arch/x86/crypto/sha256_ssse3_glue.c | 39 +++ crypto/shash.c | 58 +++++ crypto/testmgr.c | 73 +++++- drivers/md/dm-verity-fec.c | 19 +- drivers/md/dm-verity-fec.h | 5 +- drivers/md/dm-verity-target.c | 192 +++++++++++---- drivers/md/dm-verity.h | 34 +-- fs/verity/fsverity_private.h | 7 + fs/verity/hash_algs.c | 8 +- fs/verity/verify.c | 169 ++++++++++--- include/crypto/hash.h | 52 +++- 14 files changed, 1224 insertions(+), 121 deletions(-)
base-commit: 9852d85ec9d492ebef56dc5f229416c925758edc
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 1/7 crypto: shash - add support for finup_mb
From: Eric Biggers <[email protected]>
Most cryptographic hash functions are serialized, in the sense that they have an internal block size and the blocks must be processed serially. (BLAKE3 is a notable exception that has tree-based hashing built-in, but all the more common choices such as the SHAs and BLAKE2 are serialized. ParallelHash and Sakura are parallel hashes based on SHA3, but SHA3 is much slower than SHA256 in software even with the ARMv8 SHA3 extension.)
This limits the performance of computing a single hash. Yet, computing multiple hashes simultaneously does not have this limitation. Modern CPUs are superscalar and often can execute independent instructions in parallel. As a result, on many modern CPUs, it is possible to hash two equal-length messages in about the same time as a single message, if all the instructions are interleaved.
Meanwhile, a very common use case for hashing in the Linux kernel is dm-verity and fs-verity. Both use a Merkle tree that has a fixed block size, usually 4096 bytes with an empty or 32-byte salt prepended. The hash algorithm is usually SHA-256. Usually, many blocks need to be hashed at a time. This is an ideal scenario for multibuffer hashing.
Linux actually used to support SHA-256 multibuffer hashing on x86_64, before it was removed by commit ab8085c130ed ("crypto: x86 - remove SHA multibuffer routines and mcryptd"). However, it was integrated with the crypto API in a weird way, where it behaved as an asynchronous hash that queued up and executed all requests on a global queue. This made it very complex, buggy, and virtually unusable.
This patch takes a new approach of just adding an API crypto_shash_finup_mb() that synchronously computes the hash of multiple equal-length messages, starting from a common state that represents the (possibly empty) common prefix shared by the messages.
The new API is part of the "shash" algorithm type, as it does not make sense in "ahash". It does a "finup" operation rather than a "digest" operation in order to support the salt that is used by dm-verity and fs-verity. The data and output buffers are provided in arrays of length @num_msgs in order to make the API itself extensible to interleaving factors other than 2. (Though, initially only 2x will actually be used. There are some platforms in which a higher factor could help, but there are significant trade-offs.)
--- crypto/shash.c | 58 +++++++++++++++++++++++++++++++++++++++++++ include/crypto/hash.h | 52 +++++++++++++++++++++++++++++++++++++- 2 files changed, 109 insertions(+), 1 deletion(-)
crypto/shash.c +58 -0
@@ -73,10 +73,57 @@ int crypto_shash_finup(struct shash_desc *desc, const u8 *data, {return crypto_shash_alg (desc->tfm)->finup (desc, data, len, out); }EXPORT_SYMBOL_GPL (crypto_shash_finup); +static noinline_for_stack int +shash_finup_mb_fallback (struct shash_desc *desc,const u8 *const data[], +unsigned int len,u8 *const outs[], +unsigned int num_msgs) +{ +struct crypto_shash *tfm = desc->tfm; +SHASH_DESC_ON_STACK (desc2, tfm); +unsigned int i; +int err; + +for (i =0 ; i < num_msgs -1 ; i++) { + desc2->tfm = tfm; +memcpy (shash_desc_ctx (desc2),shash_desc_ctx (desc), +crypto_shash_descsize (tfm)); + err =crypto_shash_finup (desc2, data[i], len, outs[i]); +if (err) +return err; + } +return crypto_shash_finup (desc, data[i], len, outs[i]); +} + +int crypto_shash_finup_mb (struct shash_desc *desc,const u8 *const data[], +unsigned int len,u8 *const outs[], +unsigned int num_msgs) +{ +struct shash_alg *alg =crypto_shash_alg (desc->tfm); +int err; + +if (num_msgs ==1 ) +return crypto_shash_finup (desc, data[0 ], len, outs[0 ]); + +if (num_msgs ==0 ) +return 0 ; + +if (WARN_ON_ONCE (num_msgs > alg->mb_max_msgs)) + goto fallback; + + err = alg->finup_mb (desc, data, len, outs, num_msgs); +if (unlikely (err == -EOPNOTSUPP )) + goto fallback; +return err; + +fallback: +return shash_finup_mb_fallback (desc, data, len, outs, num_msgs); +} +EXPORT_SYMBOL_GPL (crypto_shash_finup_mb); +static int shash_default_digest (struct shash_desc *desc,const u8 *data,unsigned int len,u8 *out) {struct shash_alg *shash =crypto_shash_alg (desc->tfm);
@@ -312,10 +359,21 @@ static int shash_prepare_alg(struct shash_alg *alg)return -EINVAL ;if ((alg->export && !alg->import) || (alg->import && !alg->export))return -EINVAL ; +if (alg->mb_max_msgs >1 ) { +if (alg->mb_max_msgs >HASH_MAX_MB_MSGS ) +return -EINVAL ; +if (!alg->finup_mb) +return -EINVAL ; + }else { +if (alg->finup_mb) +return -EINVAL ; + alg->mb_max_msgs =1 ; + } + err =hash_prepare_alg (&alg->halg);if (err)return err; base->cra_type = &crypto_shash_type;
include/crypto/hash.h +51 -1
117 lines not on this page · raw message
Signature · 1 line
2.46.2
Herbert Xu
maintainerRe: [PATCH v7 1/7] crypto: shash - add support for finup_mb
Quoted 8 lines · Eric Biggers
Linux actually used to support SHA-256 multibuffer hashing on
x86_64, before it was removed by commitab8085c130ed("crypto: x86 - remove SHA multibuffer routines and mcryptd"). However, it was integrated with the crypto API in a weird way, where it behaved as an asynchronous hash that queued up and executed all requests on a global queue. This made it very complex, buggy, and virtually unusable.
Quoted 4 lines
This patch takes a new approach of just adding an API
crypto_shash_finup_mb()that synchronously computes the hash of multiple equal-length messages, starting from a common state that represents the (possibly empty) common prefix shared by the messages.
I absolutely love the multi-buffer idea that you've come up with. But (as you know) I dislike the shash implementation.
I'm almost there with the ahash version of this, there's just a few corner cases I need to cover due to the addition of virt address support to ahash. Incidentally this also means that I'm going to scrap lskcipher and redo it as a normal skcipher with virt address support.
BTW, I just discovered a new application for this. The marvell/cesa driver wants to chain multiple hash requests together to reduce IRQs and improve efficiency. It was doing some ad-hoc (and possibly buggy) chaining in its queueing code to achieve this.
I think having a chained submission interface solves this in a much nicer way. The driver could then simply pick up a submission chain, transform it, and then feed it directly to the hardware.
Cheers,
Signature · 3 lines
Email: Herbert Xu <[email protected]> Home Page: http://gondor.apana.org.au/~herbert/ PGP Key: http://gondor.apana.org.au/~herbert/pubkey.txt
Eric Biggers
maintainerPATCH v7 2/7 crypto: testmgr - add tests for finup_mb
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
Update the shash self-tests to test the new finup_mb method when CONFIG_CRYPTO_MANAGER_EXTRA_TESTS=y.
--- crypto/testmgr.c | 73 +++++++++++++++++++++++++++++++++++++++++++----- 1 file changed, 66 insertions(+), 7 deletions(-)
crypto/testmgr.c +66 -7
@@ -227,10 +227,11 @@ enum flush_type {/* finalization function for hash algorithms */ enum finalization_type {FINALIZATION_TYPE_FINAL ,/* use final() */ FINALIZATION_TYPE_FINUP ,/* use finup() */ +FINALIZATION_TYPE_FINUP_MB ,/* use finup_mb() */ FINALIZATION_TYPE_DIGEST ,/* use digest() */ };/* * Whether the crypto operation will occur in-place, and if so whether the
@@ -290,10 +291,14 @@ struct test_sg_division {* the @iv_offset * @key_offset: misalignment of the key, where 0 is default alignment * @key_offset_relative_to_alignmask: if true, add the algorithm's alignmask to * the @key_offset * @finalization_type: what finalization function to use for hashes +* @multibuffer_index: random number used to generate the message index to use +* for finup_mb (when finup_mb is used). +* @multibuffer_count: random number used to generate the num_msgs parameter to +* finup_mb (when finup_mb is used). * @nosimd: execute with SIMD disabled? Requires !CRYPTO_TFM_REQ_MAY_SLEEP. * This applies to the parts of the operation that aren't controlled * individually by @nosimd_setkey or @src_divs[].nosimd. * @nosimd_setkey: set the key (if applicable) with SIMD disabled? Requires * !CRYPTO_TFM_REQ_MAY_SLEEP.
@@ -307,10 +312,12 @@ struct testvec_config {unsigned int iv_offset;unsigned int key_offset;bool iv_offset_relative_to_alignmask;bool key_offset_relative_to_alignmask;enum finalization_type finalization_type; +unsigned int multibuffer_index; +unsigned int multibuffer_count;bool nosimd;bool nosimd_setkey; };#define TESTVEC_CONFIG_NAMELEN 192
@@ -1122,19 +1129,27 @@ static void generate_random_testvec_config(struct rnd_state *rng,if (prandom_bool (rng)) { cfg->req_flags |=CRYPTO_TFM_REQ_MAY_SLEEP ; p +=scnprintf (p, end - p," may_sleep" ); } -switch (prandom_u32_below (rng,4 )) { +switch (prandom_u32_below (rng,8 )) {case 0 : +case 1 : cfg->finalization_type =FINALIZATION_TYPE_FINAL ; p +=scnprintf (p, end - p," use_final" );break ; -case 1 : +case 2 : cfg->finalization_type =FINALIZATION_TYPE_FINUP ; p +=scnprintf (p, end - p," use_finup" );break ; +case 3 : +case 4 : + cfg->finalization_type =FINALIZATION_TYPE_FINUP_MB ; + cfg->multibuffer_index =prandom_u32_state (rng); + cfg->multibuffer_count =prandom_u32_state (rng); + p +=scnprintf (p, end - p," use_finup_mb" ); +break ;default : cfg->finalization_type =FINALIZATION_TYPE_DIGEST ; p +=scnprintf (p, end - p," use_digest" );break ; }
@@ -1289,10 +1304,37 @@ static inline int check_shash_op(const char *op, int err,pr_err ("alg: shash: %s %s() failed with err %d on test vector %s, cfg=\"%s\"\n" , driver, op, err, vec_name, cfg->name);return err; } +static int do_finup_mb (struct shash_desc *desc, +const u8 *data,unsigned int len,u8 *result, +const struct testvec_config *cfg, +const struct test_sglist *tsgl) +{ +struct crypto_shash *tfm = desc->tfm; +const u8 *unused_data = tsgl->bufs[XBUFSIZE -1 ]; +u8 unused_result[HASH_MAX_DIGESTSIZE ]; +const u8 *datas[HASH_MAX_MB_MSGS ]; +u8 *outs[HASH_MAX_MB_MSGS ]; +unsigned int num_msgs; +unsigned int msg_idx; +unsigned int i; + + num_msgs =1 + (cfg->multibuffer_count %crypto_shash_mb_max_msgs (tfm)); +if (WARN_ON_ONCE (num_msgs >HASH_MAX_MB_MSGS )) +return -EINVAL ; + msg_idx = cfg->multibuffer_index % num_msgs; +for (i =0 ; i < num_msgs; i++) { + datas[i] = unused_data; + outs[i] = unused_result; + } + datas[msg_idx] = data; + outs[msg_idx] = result; +return crypto_shash_finup_mb (desc, datas, len, outs, num_msgs); +} +/* Test one hash test vector in one configuration, using the shash API */ static int test_shash_vec_cfg(const struct hash_testvec *vec, const char *vec_name, const struct testvec_config *cfg, struct shash_desc *desc,
@@ -1365,11 +1407,14 @@ static int test_shash_vec_cfg(const struct hash_testvec *vec,return -EINVAL ; } goto result_ready; } -/* Using init(), zero or more update(), then final() or finup() */ +/* +* Using init(), zero or more update(), then either final(), finup(), or +* finup_mb(). +*/ if (cfg->nosimd)crypto_disable_simd_for_test (); err =crypto_shash_init (desc);if (cfg->nosimd)
@@ -1377,28 +1422,42 @@ static int test_shash_vec_cfg(const struct hash_testvec *vec, err =check_shash_op ("init" , err, driver, vec_name, cfg);if (err)return err;for (i =0 ; i < tsgl->nents; i++) { +const u8 *data =sg_virt (&tsgl->sgl[i]); +unsigned int len = tsgl->sgl[i].length; +if (i +1 == tsgl->nents && cfg->finalization_type ==FINALIZATION_TYPE_FINUP ) {if (divs[i]->nosimd)crypto_disable_simd_for_test (); - err =crypto_shash_finup (desc,sg_virt (&tsgl->sgl[i]), - tsgl->sgl[i].length, result); + err =crypto_shash_finup (desc, data, len, result);if (divs[i]->nosimd)crypto_reenable_simd_for_test (); err =check_shash_op ("finup" , err, driver, vec_name, cfg);if (err)return err; goto result_ready; } +if (i +1 == tsgl->nents && + cfg->finalization_type ==FINALIZATION_TYPE_FINUP_MB ) { +if (divs[i]->nosimd) +crypto_disable_simd_for_test (); + err =do_finup_mb (desc, data, len, result, cfg, tsgl); +if (divs[i]->nosimd) +crypto_reenable_simd_for_test (); + err =check_shash_op ("finup_mb" , err, driver, vec_name, + cfg); +if (err) +return err; + goto result_ready; + }if (divs[i]->nosimd)crypto_disable_simd_for_test (); - err =crypto_shash_update (desc,sg_virt (&tsgl->sgl[i]), - tsgl->sgl[i].length); + err =crypto_shash_update (desc, data, len);if (divs[i]->nosimd)crypto_reenable_simd_for_test (); err =check_shash_op ("update" , err, driver, vec_name, cfg);if (err)return err;
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 3/7 crypto: x86/sha256-ni - add support for finup_mb
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
Add an implementation of finup_mb to sha256-ni, using an interleaving factor of 2. It interleaves a finup operation for two equal-length messages that share a common prefix. dm-verity and fs-verity will take advantage of this for greatly improved performance on capable CPUs.
This increases the throughput of SHA-256 hashing 4096-byte messages by the following amounts on the following CPUs:
Intel Ice Lake (server): 4%
Intel Sapphire Rapids: 38%
Intel Emerald Rapids: 38%
AMD Zen 1 (Threadripper 1950X): 84%
AMD Zen 4 (EPYC 9B14): 98%
AMD Zen 5 (Ryzen 9 9950X): 64%For now, this seems to benefit AMD more than Intel. This seems to be because current AMD CPUs support concurrent execution of the SHA-NI instructions, but unfortunately current Intel CPUs don't, except for the sha256msg2 instruction. Hopefully future Intel CPUs will support SHA-NI on more execution ports. Zen 1 supports 2 concurrent sha256rnds2, and Zen 4 supports 4 concurrent sha256rnds2, which suggests that even better performance may be achievable on Zen 4 by interleaving more than two hashes. However, doing so poses a number of trade-offs, and furthermore Zen 5 goes back to supporting "only" 2 concurrent sha256rnds2.
It's been reported that the method that achieves the highest SHA-256 throughput on Intel CPUs is actually computing 16 hashes simultaneously using AVX512. That method would be quite different to the SHA-NI method used in this patch. However, such a high interleaving factor isn't practical for the use cases being targeted in the kernel.
--- arch/x86/crypto/sha256_ni_asm.S | 368 ++++++++++++++++++++++++++++ arch/x86/crypto/sha256_ssse3_glue.c | 39 +++ 2 files changed, 407 insertions(+)
arch/x86/crypto/sha256_ni_asm.S +368 -0
@@ -172,10 +172,378 @@ SYM_TYPED_FUNC_START(sha256_ni_transform) .Ldone_hash: RET SYM_FUNC_END(sha256_ni_transform) +#undef DIGEST_PTR +#undef DATA_PTR +#undef NUM_BLKS +#undef SHA256CONSTANTS +#undef MSG +#undef STATE0 +#undef STATE1 +#undef MSG0 +#undef MSG1 +#undef MSG2 +#undef MSG3 +#undef TMP +#undef SHUF_MASK +#undef ABEF_SAVE +#undef CDGH_SAVE + +// parameters for __sha256_ni_finup2x() +#define SCTX %rdi +#define DATA1 %rsi +#define DATA2 %rdx +#define LEN %ecx +#define LEN8 %cl +#define LEN64 %rcx +#define OUT1 %r8 +#define OUT2 %r9 + +// other scalar variables +#define SHA256CONSTANTS %rax +#define COUNT %r10 +#define COUNT32 %r10d +#define FINAL_STEP %r11d + +// rbx is used as a temporary. + +#define MSG %xmm0 // sha256rnds2 implicit operand +#define STATE0_A %xmm1 +#define STATE1_A %xmm2 +#define STATE0_B %xmm3 +#define STATE1_B %xmm4 +#define TMP_A %xmm5 +#define TMP_B %xmm6 +#define MSG0_A %xmm7 +#define MSG1_A %xmm8 +#define MSG2_A %xmm9 +#define MSG3_A %xmm10 +#define MSG0_B %xmm11 +#define MSG1_B %xmm12 +#define MSG2_B %xmm13 +#define MSG3_B %xmm14 +#define SHUF_MASK %xmm15 + +#define OFFSETOF_STATE 0 // offsetof(struct sha256_state, state) +#define OFFSETOF_COUNT 32 // offsetof(struct sha256_state, count) +#define OFFSETOF_BUF 40 // offsetof(struct sha256_state, buf) + +// Do 4 rounds of SHA-256 for each of two messages (interleaved). m0_a and m0_b +// contain the current 4 message schedule words for the first and second message +// respectively. +// +// If not all the message schedule words have been computed yet, then this also +// computes 4 more message schedule words for each message. m1_a-m3_a contain +// the next 3 groups of 4 message schedule words for the first message, and +// likewise m1_b-m3_b for the second. After consuming the current value of +// m0_a, this macro computes the group after m3_a and writes it to m0_a, and +// likewise for *_b. This means that the next (m0_a, m1_a, m2_a, m3_a) is the +// current (m1_a, m2_a, m3_a, m0_a), and likewise for *_b, so the caller must +// cycle through the registers accordingly. +.macro do_4rounds_2x i, m0_a, m1_a, m2_a, m3_a, m0_b, m1_b, m2_b, m3_b + movdqa (\i-32)*4(SHA256CONSTANTS), TMP_A + movdqa TMP_A, TMP_B + paddd \m0_a, TMP_A + paddd \m0_b, TMP_B +.if \i < 48 + sha256msg1 \m1_a, \m0_a + sha256msg1 \m1_b, \m0_b +.endif + movdqa TMP_A, MSG + sha256rnds2 STATE0_A, STATE1_A + movdqa TMP_B, MSG + sha256rnds2 STATE0_B, STATE1_B + pshufd $0x0E, TMP_A, MSG + sha256rnds2 STATE1_A, STATE0_A + pshufd $0x0E, TMP_B, MSG + sha256rnds2 STATE1_B, STATE0_B +.if \i < 48 + movdqa \m3_a, TMP_A + movdqa \m3_b, TMP_B + palignr $4, \m2_a, TMP_A + palignr $4, \m2_b, TMP_B + paddd TMP_A, \m0_a + paddd TMP_B, \m0_b + sha256msg2 \m3_a, \m0_a + sha256msg2 \m3_b, \m0_b +.endif +.endm + +// +// void __sha256_ni_finup2x(const struct sha256_state *sctx, +// const u8 *data1, const u8 *data2, int len, +// u8 out1[SHA256_DIGEST_SIZE], +// u8 out2[SHA256_DIGEST_SIZE]); +// +// This function computes the SHA-256 digests of two messages |data1| and +// |data2| that are both |len| bytes long, starting from the initial state +// |sctx|. |len| must be at least SHA256_BLOCK_SIZE. +// +// The instructions for the two SHA-256 operations are interleaved. On many +// CPUs, this is almost twice as fast as hashing each message individually due +// to taking better advantage of the CPU's SHA-256 and SIMD throughput. +// +SYM_FUNC_START(__sha256_ni_finup2x) + // Allocate 128 bytes of stack space, 16-byte aligned. + push %rbx + push %rbp + mov %rsp, %rbp + sub $128, %rsp + and $~15, %rsp + + // Load the shuffle mask for swapping the endianness of 32-bit words. + movdqa PSHUFFLE_BYTE_FLIP_MASK(%rip), SHUF_MASK + + // Set up pointer to the round constants. + lea K256+32*4(%rip), SHA256CONSTANTS + + // Initially we're not processing the final blocks. + xor FINAL_STEP, FINAL_STEP + + // Load the initial state from sctx->state. + movdqu OFFSETOF_STATE+0*16(SCTX), STATE0_A // DCBA + movdqu OFFSETOF_STATE+1*16(SCTX), STATE1_A // HGFE + movdqa STATE0_A, TMP_A + punpcklqdq STATE1_A, STATE0_A // FEBA + punpckhqdq TMP_A, STATE1_A // DCHG + pshufd $0x1B, STATE0_A, STATE0_A // ABEF + pshufd $0xB1, STATE1_A, STATE1_A // CDGH + + // Load sctx->count. Take the mod 64 of it to get the number of bytes + // that are buffered in sctx->buf. Also save it in a register with LEN + // added to it. + mov LEN, LEN + mov OFFSETOF_COUNT(SCTX), %rbx + lea (%rbx, LEN64, 1), COUNT + and $63, %ebx + jz .Lfinup2x_enter_loop // No bytes buffered? + + // %ebx bytes (1 to 63) are currently buffered in sctx->buf. Load them + // followed by the first 64 - %ebx bytes of data. Since LEN >= 64, we + // just load 64 bytes from each of sctx->buf, DATA1, and DATA2 + // unconditionally and rearrange the data as needed. + + movdqu OFFSETOF_BUF+0*16(SCTX), MSG0_A + movdqu OFFSETOF_BUF+1*16(SCTX), MSG1_A + movdqu OFFSETOF_BUF+2*16(SCTX), MSG2_A + movdqu OFFSETOF_BUF+3*16(SCTX), MSG3_A + movdqa MSG0_A, 0*16(%rsp) + movdqa MSG1_A, 1*16(%rsp) + movdqa MSG2_A, 2*16(%rsp) + movdqa MSG3_A, 3*16(%rsp) + + movdqu 0*16(DATA1), MSG0_A + movdqu 1*16(DATA1), MSG1_A + movdqu 2*16(DATA1), MSG2_A + movdqu 3*16(DATA1), MSG3_A + movdqu MSG0_A, 0*16(%rsp,%rbx) + movdqu MSG1_A, 1*16(%rsp,%rbx) + movdqu MSG2_A, 2*16(%rsp,%rbx) + movdqu MSG3_A, 3*16(%rsp,%rbx) + movdqa 0*16(%rsp), MSG0_A + movdqa 1*16(%rsp), MSG1_A + movdqa 2*16(%rsp), MSG2_A + movdqa 3*16(%rsp), MSG3_A + + movdqu 0*16(DATA2), MSG0_B + movdqu 1*16(DATA2), MSG1_B + movdqu 2*16(DATA2), MSG2_B + movdqu 3*16(DATA2), MSG3_B + movdqu MSG0_B, 0*16(%rsp,%rbx) + movdqu MSG1_B, 1*16(%rsp,%rbx) + movdqu MSG2_B, 2*16(%rsp,%rbx) + movdqu MSG3_B, 3*16(%rsp,%rbx) + movdqa 0*16(%rsp), MSG0_B + movdqa 1*16(%rsp), MSG1_B + movdqa 2*16(%rsp), MSG2_B + movdqa 3*16(%rsp), MSG3_B + + sub $64, %rbx // rbx = buffered - 64 + sub %rbx, DATA1 // DATA1 += 64 - buffered + sub %rbx, DATA2 // DATA2 += 64 - buffered + add %ebx, LEN // LEN += buffered - 64 + movdqa STATE0_A, STATE0_B + movdqa STATE1_A, STATE1_B + jmp .Lfinup2x_loop_have_data + +.Lfinup2x_enter_loop: + sub $64, LEN + movdqa STATE0_A, STATE0_B + movdqa STATE1_A, STATE1_B +.Lfinup2x_loop: + // Load the next two data blocks. + movdqu 0*16(DATA1), MSG0_A + movdqu 0*16(DATA2), MSG0_B + movdqu 1*16(DATA1), MSG1_A + movdqu 1*16(DATA2), MSG1_B + movdqu 2*16(DATA1), MSG2_A + movdqu 2*16(DATA2), MSG2_B + movdqu 3*16(DATA1), MSG3_A + movdqu 3*16(DATA2), MSG3_B + add $64, DATA1 + add $64, DATA2 +.Lfinup2x_loop_have_data: + // Convert the words of the data blocks from big endian. + pshufb SHUF_MASK, MSG0_A + pshufb SHUF_MASK, MSG0_B + pshufb SHUF_MASK, MSG1_A + pshufb SHUF_MASK, MSG1_B + pshufb SHUF_MASK, MSG2_A + pshufb SHUF_MASK, MSG2_B + pshufb SHUF_MASK, MSG3_A + pshufb SHUF_MASK, MSG3_B +.Lfinup2x_loop_have_bswapped_data: + + // Save the original state for each block. + movdqa STATE0_A, 0*16(%rsp) + movdqa STATE0_B, 1*16(%rsp) + movdqa STATE1_A, 2*16(%rsp) + movdqa STATE1_B, 3*16(%rsp) + + // Do the SHA-256 rounds on each block. +.irp i, 0, 16, 32, 48 + do_4rounds_2x (\i + 0), MSG0_A, MSG1_A, MSG2_A, MSG3_A, \ + MSG0_B, MSG1_B, MSG2_B, MSG3_B + do_4rounds_2x (\i + 4), MSG1_A, MSG2_A, MSG3_A, MSG0_A, \ + MSG1_B, MSG2_B, MSG3_B, MSG0_B + do_4rounds_2x (\i + 8), MSG2_A, MSG3_A, MSG0_A, MSG1_A, \ + MSG2_B, MSG3_B, MSG0_B, MSG1_B + do_4rounds_2x (\i + 12), MSG3_A, MSG0_A, MSG1_A, MSG2_A, \ + MSG3_B, MSG0_B, MSG1_B, MSG2_B +.endr + + // Add the original state for each block. + paddd 0*16(%rsp), STATE0_A + paddd 1*16(%rsp), STATE0_B + paddd 2*16(%rsp), STATE1_A + paddd 3*16(%rsp), STATE1_B + + // Update LEN and loop back if more blocks remain. + sub $64, LEN + jge .Lfinup2x_loop + + // Check if any final blocks need to be handled. + // FINAL_STEP = 2: all done + // FINAL_STEP = 1: need to do count-only padding block + // FINAL_STEP = 0: need to do the block with 0x80 padding byte + cmp $1, FINAL_STEP + jg .Lfinup2x_done + je .Lfinup2x_finalize_countonly + add $64, LEN + jz .Lfinup2x_finalize_blockaligned + + // Not block-aligned; 1 <= LEN <= 63 data bytes remain. Pad the block. + // To do this, write the padding starting with the 0x80 byte to + // &sp[64]. Then for each message, copy the last 64 data bytes to sp + // and load from &sp[64 - LEN] to get the needed padding block. This + // code relies on the data buffers being >= 64 bytes in length. + mov $64, %ebx + sub LEN, %ebx // ebx = 64 - LEN + sub %rbx, DATA1 // DATA1 -= 64 - LEN + sub %rbx, DATA2 // DATA2 -= 64 - LEN + mov $0x80, FINAL_STEP // using FINAL_STEP as a temporary + movd FINAL_STEP, MSG0_A + pxor MSG1_A, MSG1_A + movdqa MSG0_A, 4*16(%rsp) + movdqa MSG1_A, 5*16(%rsp) + movdqa MSG1_A, 6*16(%rsp) + movdqa MSG1_A, 7*16(%rsp) + cmp $56, LEN + jge 1f // will COUNT spill into its own block? + shl $3, COUNT + bswap COUNT + mov COUNT, 56(%rsp,%rbx) + mov $2, FINAL_STEP // won't need count-only block + jmp 2f +1: + mov $1, FINAL_STEP // will need count-only block +2: + movdqu 0*16(DATA1), MSG0_A + movdqu 1*16(DATA1), MSG1_A + movdqu 2*16(DATA1), MSG2_A + movdqu 3*16(DATA1), MSG3_A + movdqa MSG0_A, 0*16(%rsp) + movdqa MSG1_A, 1*16(%rsp) + movdqa MSG2_A, 2*16(%rsp) + movdqa MSG3_A, 3*16(%rsp) + movdqu 0*16(%rsp,%rbx), MSG0_A + movdqu 1*16(%rsp,%rbx), MSG1_A + movdqu 2*16(%rsp,%rbx), MSG2_A + movdqu 3*16(%rsp,%rbx), MSG3_A + + movdqu 0*16(DATA2), MSG0_B + movdqu 1*16(DATA2), MSG1_B + movdqu 2*16(DATA2), MSG2_B + movdqu 3*16(DATA2), MSG3_B + movdqa MSG0_B, 0*16(%rsp) + movdqa MSG1_B, 1*16(%rsp) + movdqa MSG2_B, 2*16(%rsp) + movdqa MSG3_B, 3*16(%rsp) + movdqu 0*16(%rsp,%rbx), MSG0_B + movdqu 1*16(%rsp,%rbx), MSG1_B + movdqu 2*16(%rsp,%rbx), MSG2_B + movdqu 3*16(%rsp,%rbx), MSG3_B + jmp .Lfinup2x_loop_have_data + + // Prepare a padding block, either: + // + // {0x80, 0, 0, 0, ..., count (as __be64)} + // This is for a block aligned message. + // + // { 0, 0, 0, 0, ..., count (as __be64)} + // This is for a message whose length mod 64 is >= 56. + // + // Pre-swap the endianness of the words. +.Lfinup2x_finalize_countonly: + pxor MSG0_A, MSG0_A + jmp 1f + +.Lfinup2x_finalize_blockaligned: + mov $0x80000000, %ebx + movd %ebx, MSG0_A +1: + pxor MSG1_A, MSG1_A + pxor MSG2_A, MSG2_A + ror $29, COUNT + movq COUNT, MSG3_A + pslldq $8, MSG3_A + movdqa MSG0_A, MSG0_B + pxor MSG1_B, MSG1_B + pxor MSG2_B, MSG2_B + movdqa MSG3_A, MSG3_B + mov $2, FINAL_STEP + jmp .Lfinup2x_loop_have_bswapped_data + +.Lfinup2x_done: + // Write the two digests with all bytes in the correct order. + movdqa STATE0_A, TMP_A + movdqa STATE0_B, TMP_B + punpcklqdq STATE1_A, STATE0_A // GHEF + punpcklqdq STATE1_B, STATE0_B + punpckhqdq TMP_A, STATE1_A // ABCD + punpckhqdq TMP_B, STATE1_B + pshufd $0xB1, STATE0_A, STATE0_A // HGFE + pshufd $0xB1, STATE0_B, STATE0_B + pshufd $0x1B, STATE1_A, STATE1_A // DCBA + pshufd $0x1B, STATE1_B, STATE1_B + pshufb SHUF_MASK, STATE0_A + pshufb SHUF_MASK, STATE0_B + pshufb SHUF_MASK, STATE1_A + pshufb SHUF_MASK, STATE1_B + movdqu STATE0_A, 1*16(OUT1) + movdqu STATE0_B, 1*16(OUT2) + movdqu STATE1_A, 0*16(OUT1) + movdqu STATE1_B, 0*16(OUT2) + + mov %rbp, %rsp + pop %rbp + pop %rbx + RET +SYM_FUNC_END(__sha256_ni_finup2x) + .section .rodata.cst256.K256, "aM", @progbits, 256 .align 64 K256: .long 0x428a2f98,0x71374491,0xb5c0fbcf,0xe9b5dba5 .long 0x3956c25b,0x59f111f1,0x923f82a4,0xab1c5ed5
arch/x86/crypto/sha256_ssse3_glue.c +39 -0
@@ -331,10 +331,15 @@ static void unregister_sha256_avx2(void)#ifdef CONFIG_AS_SHA256_NI asmlinkage void sha256_ni_transform(struct sha256_state *digest,const u8 *data,int rounds); +asmlinkage void __sha256_ni_finup2x(const struct sha256_state *sctx, +const u8 *data1,const u8 *data2,int len, +u8 out1[SHA256_DIGEST_SIZE ], +u8 out2[SHA256_DIGEST_SIZE ]); +static int sha256_ni_update (struct shash_desc *desc,const u8 *data,unsigned int len) {return _sha256_update (desc, data, len, sha256_ni_transform); }
@@ -355,18 +360,52 @@ static int sha256_ni_digest(struct shash_desc *desc, const u8 *data, {return sha256_base_init (desc) ?:sha256_ni_finup (desc, data, len, out); } +static int sha256_ni_finup_mb (struct shash_desc *desc, +const u8 *const data[],unsigned int len, +u8 *const outs[],unsigned int num_msgs) +{ +struct sha256_state *sctx =shash_desc_ctx (desc); + +/* +* num_msgs != 2 should not happen here, since this algorithm sets +* mb_max_msgs=2, and the crypto API handles num_msgs <= 1 before +* calling into the algorithm's finup_mb method. +*/ +if (WARN_ON_ONCE (num_msgs !=2 )) +return -EOPNOTSUPP ; + +if (unlikely (!crypto_simd_usable ())) +return -EOPNOTSUPP ; + +/* __sha256_ni_finup2x() assumes SHA256_BLOCK_SIZE <= len <= INT_MAX. */ +if (unlikely (len <SHA256_BLOCK_SIZE || len >INT_MAX )) +return -EOPNOTSUPP ; + +/* __sha256_ni_finup2x() assumes the following offsets. */ +BUILD_BUG_ON (offsetof(struct sha256_state , state) !=0 ); +BUILD_BUG_ON (offsetof(struct sha256_state , count) !=32 ); +BUILD_BUG_ON (offsetof(struct sha256_state , buf) !=40 ); + +kernel_fpu_begin (); +__sha256_ni_finup2x (sctx, data[0 ], data[1 ], len, outs[0 ], outs[1 ]); +kernel_fpu_end (); +return 0 ; +} +static struct shash_alg sha256_ni_algs[] = { { .digestsize =SHA256_DIGEST_SIZE , .init = sha256_base_init, .update = sha256_ni_update, .final = sha256_ni_final, .finup = sha256_ni_finup, .digest = sha256_ni_digest, + .finup_mb = sha256_ni_finup_mb, .descsize =sizeof (struct sha256_state ), + .mb_max_msgs =2 , .base = { .cra_name ="sha256" , .cra_driver_name ="sha256-ni" , .cra_priority =250 , .cra_blocksize =SHA256_BLOCK_SIZE ,
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 4/7 crypto: arm64/sha256-ce - add support for finup_mb
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
Add an implementation of finup_mb to sha256-ce, using an interleaving factor of 2. It interleaves a finup operation for two equal-length messages that share a common prefix. dm-verity and fs-verity will take advantage of this for greatly improved performance on capable CPUs.
This increases the throughput of SHA-256 hashing 4096-byte messages by the following amounts on the following CPUs:
ARM Cortex-X1: 70%
ARM Cortex-X3: 68%
ARM Cortex-A76: 65%
ARM Cortex-A715: 43%
ARM Cortex-A510: 25%
ARM Cortex-A55: 8%--- arch/arm64/crypto/sha2-ce-core.S | 281 ++++++++++++++++++++++++++++++- arch/arm64/crypto/sha2-ce-glue.c | 40 +++++ 2 files changed, 315 insertions(+), 6 deletions(-)
arch/arm64/crypto/sha2-ce-core.S +275 -6
@@ -68,22 +68,26 @@ .word 0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5 .word 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3 .word 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208 .word 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2 + .macro load_round_constants tmp + adr_l \tmp, .Lsha2_rcon + ld1 { v0.4s- v3.4s}, [\tmp], #64 + ld1 { v4.4s- v7.4s}, [\tmp], #64 + ld1 { v8.4s-v11.4s}, [\tmp], #64 + ld1 {v12.4s-v15.4s}, [\tmp] + .endm + /* * int __sha256_ce_transform(struct sha256_ce_state *sst, u8 const *src, * int blocks) */ .text SYM_FUNC_START(__sha256_ce_transform) - /* load round constants */ - adr_l x8, .Lsha2_rcon - ld1 { v0.4s- v3.4s}, [x8], #64 - ld1 { v4.4s- v7.4s}, [x8], #64 - ld1 { v8.4s-v11.4s}, [x8], #64 - ld1 {v12.4s-v15.4s}, [x8] + + load_round_constants x8 /* load state */ ld1 {dgav.4s, dgbv.4s}, [x0] /* load sha256_ce_state::finalize */
@@ -153,5 +157,270 @@ CPU_LE( rev32 v19.16b, v19.16b ) /* store new state */ 3: st1 {dgav.4s, dgbv.4s}, [x0] mov w0, w2 ret SYM_FUNC_END(__sha256_ce_transform) + + .unreq dga + .unreq dgav + .unreq dgb + .unreq dgbv + .unreq t0 + .unreq t1 + .unreq dg0q + .unreq dg0v + .unreq dg1q + .unreq dg1v + .unreq dg2q + .unreq dg2v + + // parameters for __sha256_ce_finup2x() + sctx .req x0 + data1 .req x1 + data2 .req x2 + len .req w3 + out1 .req x4 + out2 .req x5 + + // other scalar variables + count .req x6 + final_step .req w7 + + // x8-x9 are used as temporaries. + + // v0-v15 are used to cache the SHA-256 round constants. + // v16-v19 are used for the message schedule for the first message. + // v20-v23 are used for the message schedule for the second message. + // v24-v31 are used for the state and temporaries as given below. + // *_a are for the first message and *_b for the second. + state0_a_q .req q24 + state0_a .req v24 + state1_a_q .req q25 + state1_a .req v25 + state0_b_q .req q26 + state0_b .req v26 + state1_b_q .req q27 + state1_b .req v27 + t0_a .req v28 + t0_b .req v29 + t1_a_q .req q30 + t1_a .req v30 + t1_b_q .req q31 + t1_b .req v31 + +#define OFFSETOF_COUNT 32 // offsetof(struct sha256_state, count) +#define OFFSETOF_BUF 40 // offsetof(struct sha256_state, buf) +// offsetof(struct sha256_state, state) is assumed to be 0. + + // Do 4 rounds of SHA-256 for each of two messages (interleaved). m0_a + // and m0_b contain the current 4 message schedule words for the first + // and second message respectively. + // + // If not all the message schedule words have been computed yet, then + // this also computes 4 more message schedule words for each message. + // m1_a-m3_a contain the next 3 groups of 4 message schedule words for + // the first message, and likewise m1_b-m3_b for the second. After + // consuming the current value of m0_a, this macro computes the group + // after m3_a and writes it to m0_a, and likewise for *_b. This means + // that the next (m0_a, m1_a, m2_a, m3_a) is the current (m1_a, m2_a, + // m3_a, m0_a), and likewise for *_b, so the caller must cycle through + // the registers accordingly. + .macro do_4rounds_2x i, k, m0_a, m1_a, m2_a, m3_a, \ + m0_b, m1_b, m2_b, m3_b + add t0_a\().4s, \m0_a\().4s, \k\().4s + add t0_b\().4s, \m0_b\().4s, \k\().4s + .if \i < 48 + sha256su0 \m0_a\().4s, \m1_a\().4s + sha256su0 \m0_b\().4s, \m1_b\().4s + sha256su1 \m0_a\().4s, \m2_a\().4s, \m3_a\().4s + sha256su1 \m0_b\().4s, \m2_b\().4s, \m3_b\().4s + .endif + mov t1_a.16b, state0_a.16b + mov t1_b.16b, state0_b.16b + sha256h state0_a_q, state1_a_q, t0_a\().4s + sha256h state0_b_q, state1_b_q, t0_b\().4s + sha256h2 state1_a_q, t1_a_q, t0_a\().4s + sha256h2 state1_b_q, t1_b_q, t0_b\().4s + .endm + + .macro do_16rounds_2x i, k0, k1, k2, k3 + do_4rounds_2x \i + 0, \k0, v16, v17, v18, v19, v20, v21, v22, v23 + do_4rounds_2x \i + 4, \k1, v17, v18, v19, v16, v21, v22, v23, v20 + do_4rounds_2x \i + 8, \k2, v18, v19, v16, v17, v22, v23, v20, v21 + do_4rounds_2x \i + 12, \k3, v19, v16, v17, v18, v23, v20, v21, v22 + .endm + +// +// void __sha256_ce_finup2x(const struct sha256_state *sctx, +// const u8 *data1, const u8 *data2, int len, +// u8 out1[SHA256_DIGEST_SIZE], +// u8 out2[SHA256_DIGEST_SIZE]); +// +// This function computes the SHA-256 digests of two messages |data1| and +// |data2| that are both |len| bytes long, starting from the initial state +// |sctx|. |len| must be at least SHA256_BLOCK_SIZE. +// +// The instructions for the two SHA-256 operations are interleaved. On many +// CPUs, this is almost twice as fast as hashing each message individually due +// to taking better advantage of the CPU's SHA-256 and SIMD throughput. +// +SYM_FUNC_START(__sha256_ce_finup2x) + sub sp, sp, #128 + mov final_step, #0 + load_round_constants x8 + + // Load the initial state from sctx->state. + ld1 {state0_a.4s-state1_a.4s}, [sctx] + + // Load sctx->count. Take the mod 64 of it to get the number of bytes + // that are buffered in sctx->buf. Also save it in a register with len + // added to it. + ldr x8, [sctx, #OFFSETOF_COUNT] + add count, x8, len, sxtw + and x8, x8, #63 + cbz x8, .Lfinup2x_enter_loop // No bytes buffered? + + // x8 bytes (1 to 63) are currently buffered in sctx->buf. Load them + // followed by the first 64 - x8 bytes of data. Since len >= 64, we + // just load 64 bytes from each of sctx->buf, data1, and data2 + // unconditionally and rearrange the data as needed. + add x9, sctx, #OFFSETOF_BUF + ld1 {v16.16b-v19.16b}, [x9] + st1 {v16.16b-v19.16b}, [sp] + + ld1 {v16.16b-v19.16b}, [data1], #64 + add x9, sp, x8 + st1 {v16.16b-v19.16b}, [x9] + ld1 {v16.4s-v19.4s}, [sp] + + ld1 {v20.16b-v23.16b}, [data2], #64 + st1 {v20.16b-v23.16b}, [x9] + ld1 {v20.4s-v23.4s}, [sp] + + sub len, len, #64 + sub data1, data1, x8 + sub data2, data2, x8 + add len, len, w8 + mov state0_b.16b, state0_a.16b + mov state1_b.16b, state1_a.16b + b .Lfinup2x_loop_have_data + +.Lfinup2x_enter_loop: + sub len, len, #64 + mov state0_b.16b, state0_a.16b + mov state1_b.16b, state1_a.16b +.Lfinup2x_loop: + // Load the next two data blocks. + ld1 {v16.4s-v19.4s}, [data1], #64 + ld1 {v20.4s-v23.4s}, [data2], #64 +.Lfinup2x_loop_have_data: + // Convert the words of the data blocks from big endian. +CPU_LE( rev32 v16.16b, v16.16b ) +CPU_LE( rev32 v17.16b, v17.16b ) +CPU_LE( rev32 v18.16b, v18.16b ) +CPU_LE( rev32 v19.16b, v19.16b ) +CPU_LE( rev32 v20.16b, v20.16b ) +CPU_LE( rev32 v21.16b, v21.16b ) +CPU_LE( rev32 v22.16b, v22.16b ) +CPU_LE( rev32 v23.16b, v23.16b ) +.Lfinup2x_loop_have_bswapped_data: + + // Save the original state for each block. + st1 {state0_a.4s-state1_b.4s}, [sp] + + // Do the SHA-256 rounds on each block. + do_16rounds_2x 0, v0, v1, v2, v3 + do_16rounds_2x 16, v4, v5, v6, v7 + do_16rounds_2x 32, v8, v9, v10, v11 + do_16rounds_2x 48, v12, v13, v14, v15 + + // Add the original state for each block. + ld1 {v16.4s-v19.4s}, [sp] + add state0_a.4s, state0_a.4s, v16.4s + add state1_a.4s, state1_a.4s, v17.4s + add state0_b.4s, state0_b.4s, v18.4s + add state1_b.4s, state1_b.4s, v19.4s + + // Update len and loop back if more blocks remain. + sub len, len, #64 + tbz len, #31, .Lfinup2x_loop // len >= 0? + + // Check if any final blocks need to be handled. + // final_step = 2: all done + // final_step = 1: need to do count-only padding block + // final_step = 0: need to do the block with 0x80 padding byte + tbnz final_step, #1, .Lfinup2x_done + tbnz final_step, #0, .Lfinup2x_finalize_countonly + add len, len, #64 + cbz len, .Lfinup2x_finalize_blockaligned + + // Not block-aligned; 1 <= len <= 63 data bytes remain. Pad the block. + // To do this, write the padding starting with the 0x80 byte to + // &sp[64]. Then for each message, copy the last 64 data bytes to sp + // and load from &sp[64 - len] to get the needed padding block. This + // code relies on the data buffers being >= 64 bytes in length. + sub w8, len, #64 // w8 = len - 64 + add data1, data1, w8, sxtw // data1 += len - 64 + add data2, data2, w8, sxtw // data2 += len - 64 + mov x9, 0x80 + fmov d16, x9 + movi v17.16b, #0 + stp q16, q17, [sp, #64] + stp q17, q17, [sp, #96] + sub x9, sp, w8, sxtw // x9 = &sp[64 - len] + cmp len, #56 + b.ge 1f // will count spill into its own block? + lsl count, count, #3 + rev count, count + str count, [x9, #56] + mov final_step, #2 // won't need count-only block + b 2f +1: + mov final_step, #1 // will need count-only block +2: + ld1 {v16.16b-v19.16b}, [data1] + st1 {v16.16b-v19.16b}, [sp] + ld1 {v16.4s-v19.4s}, [x9] + ld1 {v20.16b-v23.16b}, [data2] + st1 {v20.16b-v23.16b}, [sp] + ld1 {v20.4s-v23.4s}, [x9] + b .Lfinup2x_loop_have_data + + // Prepare a padding block, either: + // + // {0x80, 0, 0, 0, ..., count (as __be64)} + // This is for a block aligned message. + // + // { 0, 0, 0, 0, ..., count (as __be64)} + // This is for a message whose length mod 64 is >= 56. + // + // Pre-swap the endianness of the words. +.Lfinup2x_finalize_countonly: + movi v16.2d, #0 + b 1f +.Lfinup2x_finalize_blockaligned: + mov x8, #0x80000000 + fmov d16, x8 +1: + movi v17.2d, #0 + movi v18.2d, #0 + ror count, count, #29 // ror(lsl(count, 3), 32) + mov v19.d[0], xzr + mov v19.d[1], count + mov v20.16b, v16.16b + movi v21.2d, #0 + movi v22.2d, #0 + mov v23.16b, v19.16b + mov final_step, #2 + b .Lfinup2x_loop_have_bswapped_data + +.Lfinup2x_done: + // Write the two digests with all bytes in the correct order. +CPU_LE( rev32 state0_a.16b, state0_a.16b ) +CPU_LE( rev32 state1_a.16b, state1_a.16b ) +CPU_LE( rev32 state0_b.16b, state0_b.16b ) +CPU_LE( rev32 state1_b.16b, state1_b.16b ) + st1 {state0_a.4s-state1_a.4s}, [out1] + st1 {state0_b.4s-state1_b.4s}, [out2] + add sp, sp, #128 + ret +SYM_FUNC_END(__sha256_ce_finup2x)
arch/arm64/crypto/sha2-ce-glue.c +40 -0
@@ -31,10 +31,15 @@ extern const u32 sha256_ce_offsetof_count;extern const u32 sha256_ce_offsetof_finalize; asmlinkage int __sha256_ce_transform(struct sha256_ce_state *sst, u8 const *src,int blocks); +asmlinkage void __sha256_ce_finup2x(const struct sha256_state *sctx, +const u8 *data1,const u8 *data2,int len, +u8 out1[SHA256_DIGEST_SIZE ], +u8 out2[SHA256_DIGEST_SIZE ]); +static void sha256_ce_transform (struct sha256_state *sst,u8 const *src,int blocks) {while (blocks) {int rem;
@@ -122,10 +127,43 @@ static int sha256_ce_digest(struct shash_desc *desc, const u8 *data, {sha256_base_init (desc);return sha256_ce_finup (desc, data, len, out); } +static int sha256_ce_finup_mb (struct shash_desc *desc, +const u8 *const data[],unsigned int len, +u8 *const outs[],unsigned int num_msgs) +{ +struct sha256_ce_state *sctx =shash_desc_ctx (desc); + +/* +* num_msgs != 2 should not happen here, since this algorithm sets +* mb_max_msgs=2, and the crypto API handles num_msgs <= 1 before +* calling into the algorithm's finup_mb method. +*/ +if (WARN_ON_ONCE (num_msgs !=2 )) +return -EOPNOTSUPP ; + +if (unlikely (!crypto_simd_usable ())) +return -EOPNOTSUPP ; + +/* __sha256_ce_finup2x() assumes SHA256_BLOCK_SIZE <= len <= INT_MAX. */ +if (unlikely (len <SHA256_BLOCK_SIZE || len >INT_MAX )) +return -EOPNOTSUPP ; + +/* __sha256_ce_finup2x() assumes the following offsets. */ +BUILD_BUG_ON (offsetof(struct sha256_state , state) !=0 ); +BUILD_BUG_ON (offsetof(struct sha256_state , count) !=32 ); +BUILD_BUG_ON (offsetof(struct sha256_state , buf) !=40 ); + +kernel_neon_begin (); +__sha256_ce_finup2x (&sctx->sst, data[0 ], data[1 ], len, outs[0 ], + outs[1 ]); +kernel_neon_end (); +return 0 ; +} +static int sha256_ce_export (struct shash_desc *desc,void *out) {struct sha256_ce_state *sctx =shash_desc_ctx (desc);memcpy (out, &sctx->sst,sizeof (struct sha256_state ));
@@ -162,13 +200,15 @@ static struct shash_alg algs[] = { { .init = sha256_base_init, .update = sha256_ce_update, .final = sha256_ce_final, .finup = sha256_ce_finup, .digest = sha256_ce_digest, + .finup_mb = sha256_ce_finup_mb, .export = sha256_ce_export, .import = sha256_ce_import, .descsize = sizeof(struct sha256_ce_state), + .mb_max_msgs = 2, .statesize = sizeof(struct sha256_state), .digestsize = SHA256_DIGEST_SIZE, .base = { .cra_name = "sha256", .cra_driver_name = "sha256-ce",
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 5/7 fsverity: improve performance by using multibuffer hashing
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
When supported by the hash algorithm, use crypto_shash_finup_mb() to interleave the hashing of pairs of data blocks. On some CPUs this nearly doubles hashing performance. The increase in overall throughput of cold-cache fsverity reads that I'm seeing on arm64 and x86_64 is roughly 35% (though this metric is hard to measure as it jumps around a lot).
For now this is only done on the verification path, and only for data blocks, not Merkle tree blocks. We could use finup_mb on Merkle tree blocks too, but that is less important as there aren't as many Merkle tree blocks as data blocks, and that would require some additional code restructuring. We could also use finup_mb to accelerate building the Merkle tree, but verification performance is more important.
--- fs/verity/fsverity_private.h | 7 ++ fs/verity/hash_algs.c | 8 +- fs/verity/verify.c | 169 +++++++++++++++++++++++++++++------ 3 files changed, 153 insertions(+), 31 deletions(-)
fs/verity/fsverity_private.h +7 -0
@@ -27,10 +27,15 @@ struct fsverity_hash_alg {/* * The HASH_ALGO_* constant for this algorithm. This is different from * FS_VERITY_HASH_ALG_*, which uses a different numbering scheme. */ enum hash_algo algo_id; +/* +* The maximum supported interleaving factor for multibuffer hashing, or +* 1 if the algorithm doesn't support multibuffer hashing +*/ +int mb_max_msgs; };/* Merkle tree parameters: hash algorithm, initial hash state, and topology */ struct merkle_tree_params {const struct fsverity_hash_alg *hash_alg;/* the hash algorithm */
@@ -150,8 +155,10 @@ static inline void fsverity_init_signature(void) }#endif /* !CONFIG_FS_VERITY_BUILTIN_SIGNATURES */ /* verify.c */ +#define FS_VERITY_MAX_PENDING_DATA_BLOCKS 2 +void __init fsverity_init_workqueue (void );#endif /* _FSVERITY_PRIVATE_H */
fs/verity/hash_algs.c +6 -2
@@ -82,12 +82,16 @@ const struct fsverity_hash_alg *fsverity_get_hash_alg(const struct inode *inode,if (WARN_ON_ONCE (alg->digest_size !=crypto_shash_digestsize (tfm))) goto err_free_tfm;if (WARN_ON_ONCE (alg->block_size !=crypto_shash_blocksize (tfm))) goto err_free_tfm; -pr_info ("%s using implementation \"%s\"\n" , - alg->name,crypto_shash_driver_name (tfm)); + alg->mb_max_msgs =min (crypto_shash_mb_max_msgs (tfm), +FS_VERITY_MAX_PENDING_DATA_BLOCKS ); + +pr_info ("%s using implementation \"%s\"%s\n" , + alg->name,crypto_shash_driver_name (tfm), + alg->mb_max_msgs >1 ?" (multibuffer)" :"" );/* pairs with smp_load_acquire() above */ smp_store_release (&alg->tfm, tfm); goto out_unlock;
fs/verity/verify.c +140 -29
288 lines not on this page · raw message
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 6/7 dm-verity: reduce scope of real and wanted digests
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
In preparation for supporting multibuffer hashing where dm-verity will need to keep track of the real and wanted digests for multiple data blocks simultaneously, stop using the want_digest and real_digest fields of struct dm_verity_io from so many different places. Specifically:
- Make various functions take want_digest as a parameter rather than having it be implicitly passed via the struct dm_verity_io.
- Add a new tmp_digest field, and use this instead of real_digest when computing a hash solely for the purpose of immediately checking it.
The result is that real_digest and want_digest are used only by verity_verify_io().
--- drivers/md/dm-verity-fec.c | 19 +++++++++--------- drivers/md/dm-verity-fec.h | 5 +++-- drivers/md/dm-verity-target.c | 36 ++++++++++++++++++----------------- drivers/md/dm-verity.h | 1 + 4 files changed, 32 insertions(+), 29 deletions(-)
drivers/md/dm-verity-fec.c +9 -10
@@ -185,15 +185,14 @@ static int fec_decode_bufs(struct dm_verity *v, struct dm_verity_io *io,*/ static int fec_is_erasure (struct dm_verity *v,struct dm_verity_io *io,u8 *want_digest,u8 *data) {if (unlikely (verity_hash (v, io, data,1 << v->data_dev_block_bits, -verity_io_real_digest (v, io), true))) + io->tmp_digest, true)))return 0 ; -return memcmp (verity_io_real_digest (v, io), want_digest, - v->digest_size) !=0 ; +return memcmp (io->tmp_digest, want_digest, v->digest_size) !=0 ; }/* * Read data blocks that are part of the RS block and deinterleave as much as * fits into buffers. Check for erasure locations if @neras is non-NULL.
@@ -360,11 +359,11 @@ static void fec_init_bufs(struct dm_verity *v, struct dm_verity_fec_io *fio)* (indicated by @offset) in fio->output. If @use_erasures is non-zero, uses * hashes to locate erasures. */ static int fec_decode_rsb (struct dm_verity *v,struct dm_verity_io *io,struct dm_verity_fec_io *fio,u64 rsb,u64 offset, -bool use_erasures) +const u8 *want_digest,bool use_erasures) {int r, neras =0 ;unsigned int pos; r =fec_alloc_bufs (v, fio);
@@ -386,27 +385,27 @@ static int fec_decode_rsb(struct dm_verity *v, struct dm_verity_io *io, pos += fio->nbufs <<DM_VERITY_FEC_BUF_RS_BITS ; }/* Always re-validate the corrected block against the expected hash */ r =verity_hash (v, io, fio->output,1 << v->data_dev_block_bits, -verity_io_real_digest (v, io), true); + io->tmp_digest, true);if (unlikely (r <0 ))return r; -if (memcmp (verity_io_real_digest (v, io),verity_io_want_digest (v, io), - v->digest_size)) { +if (memcmp (io->tmp_digest, want_digest, v->digest_size)) {DMERR_LIMIT ("%s: FEC %llu: failed to correct (%d erasures)" , v->data_dev->name, (unsigned long long )rsb, neras);return -EILSEQ ; }return 0 ; }/* Correct errors in a block. Copies corrected block to dest. */ int verity_fec_decode (struct dm_verity *v,struct dm_verity_io *io, -enum verity_block_type type,sector_t block,u8 *dest) +enum verity_block_type type,const u8 *want_digest, +sector_t block,u8 *dest) {int r;struct dm_verity_fec_io *fio =fec_io (io);u64 offset, res, rsb;
@@ -445,13 +444,13 @@ int verity_fec_decode(struct dm_verity *v, struct dm_verity_io *io,/* * Locating erasures is slow, so attempt to recover the block without * them first. Do a second attempt with erasures if the corruption is * bad enough. */ - r =fec_decode_rsb (v, io, fio, rsb, offset, false); + r =fec_decode_rsb (v, io, fio, rsb, offset, want_digest, false);if (r <0 ) { - r =fec_decode_rsb (v, io, fio, rsb, offset, true); + r =fec_decode_rsb (v, io, fio, rsb, offset, want_digest, true);if (r <0 ) goto done; }memcpy (dest, fio->output,1 << v->data_dev_block_bits);
drivers/md/dm-verity-fec.h +3 -2
@@ -66,12 +66,12 @@ struct dm_verity_fec_io {#define DM_VERITY_OPTS_FEC 8extern bool verity_fec_is_enabled (struct dm_verity *v);extern int verity_fec_decode (struct dm_verity *v,struct dm_verity_io *io, -enum verity_block_type type,sector_t block, -u8 *dest); +enum verity_block_type type,const u8 *want_digest, +sector_t block,u8 *dest);extern unsigned int verity_fec_status_table (struct dm_verity *v,unsigned int sz,char *result,unsigned int maxlen);extern void verity_fec_finish_io (struct dm_verity_io *io);
@@ -97,10 +97,11 @@ static inline bool verity_fec_is_enabled(struct dm_verity *v) }static inline int verity_fec_decode (struct dm_verity *v,struct dm_verity_io *io,enum verity_block_type type, +const u8 *want_digest,sector_t block,u8 *dest) {return -EOPNOTSUPP ; }
drivers/md/dm-verity-target.c +19 -17
@@ -286,16 +286,16 @@ static int verity_handle_err(struct dm_verity *v, enum verity_block_type type,/* * Verify hash of a metadata block pertaining to the specified data block * ("block" argument) at a specified level ("level" argument). * -* On successful return, verity_io_want_digest(v, io) contains the hash value -* for a lower tree level or for the data block (if we're at the lowest level). +* On successful return, want_digest contains the hash value for a lower tree +* level or for the data block (if we're at the lowest level). * * If "skip_unverified" is true, unverified buffer is skipped and 1 is returned. * If "skip_unverified" is false, unverified buffer is hashed and verified -* against current value of verity_io_want_digest(v, io). +* against current value of want_digest. */ static int verity_verify_level (struct dm_verity *v,struct dm_verity_io *io,sector_t block,int level,bool skip_unverified,u8 *want_digest) {
@@ -334,26 +334,26 @@ static int verity_verify_level(struct dm_verity *v, struct dm_verity_io *io, r =1 ; goto release_ret_r; } r =verity_hash (v, io, data,1 << v->hash_dev_block_bits, -verity_io_real_digest (v, io), !io->in_bh); + io->tmp_digest, !io->in_bh);if (unlikely (r <0 )) goto release_ret_r; -if (likely (memcmp (verity_io_real_digest (v, io), want_digest, +if (likely (memcmp (io->tmp_digest, want_digest, v->digest_size) ==0 )) aux->hash_verified =1 ;else if (static_branch_unlikely (&use_bh_wq_enabled) && io->in_bh) {/* * Error handling code (FEC included) cannot be run in a * tasklet since it may sleep, so fallback to work-queue. */ r = -EAGAIN ; goto release_ret_r; }else if (verity_fec_decode (v, io,DM_VERITY_BLOCK_TYPE_METADATA , - hash_block, data) ==0 ) + want_digest, hash_block, data) ==0 ) aux->hash_verified =1 ;else if (verity_handle_err (v,DM_VERITY_BLOCK_TYPE_METADATA , hash_block)) { struct bio *bio =
@@ -412,11 +412,12 @@ int verity_hash_for_block(struct dm_verity *v, struct dm_verity_io *io,return r; } static noinline int verity_recheck(struct dm_verity *v, struct dm_verity_io *io, -sector_t cur_block,u8 *dest) +const u8 *want_digest,sector_t cur_block, +u8 *dest) {struct page *page;void *buffer;int r;struct dm_io_request io_req;
@@ -436,16 +437,15 @@ static noinline int verity_recheck(struct dm_verity *v, struct dm_verity_io *io, r =dm_io (&io_req,1 , &io_loc,NULL ,IOPRIO_DEFAULT );if (unlikely (r)) goto free_ret; r =verity_hash (v, io, buffer,1 << v->data_dev_block_bits, -verity_io_real_digest (v, io), true); + io->tmp_digest, true);if (unlikely (r)) goto free_ret; -if (memcmp (verity_io_real_digest (v, io), -verity_io_want_digest (v, io), v->digest_size)) { +if (memcmp (io->tmp_digest, want_digest, v->digest_size)) { r = -EIO ; goto free_ret; }memcpy (dest, buffer,1 << v->data_dev_block_bits);
@@ -456,28 +456,29 @@ static noinline int verity_recheck(struct dm_verity *v, struct dm_verity_io *io,return r; }static int verity_handle_data_hash_mismatch (struct dm_verity *v,struct dm_verity_io *io, -struct bio *bio,sector_t blkno, -u8 *data) +struct bio *bio, +const u8 *want_digest, +sector_t blkno,u8 *data) {if (static_branch_unlikely (&use_bh_wq_enabled) && io->in_bh) {/* * Error handling code (FEC included) cannot be run in the * BH workqueue, so fallback to a standard workqueue. */ return -EAGAIN ; } -if (verity_recheck (v, io, blkno, data) ==0 ) { +if (verity_recheck (v, io, want_digest, blkno, data) ==0 ) {if (v->validated_blocks)set_bit (blkno, v->validated_blocks);return 0 ; }#if defined(CONFIG_DM_VERITY_FEC ) -if (verity_fec_decode (v, io,DM_VERITY_BLOCK_TYPE_DATA , blkno, - data) ==0 ) +if (verity_fec_decode (v, io,DM_VERITY_BLOCK_TYPE_DATA , want_digest, + blkno, data) ==0 )return 0 ;#endif if (bio->bi_status)return -EIO ;/* Error correction failed; Just return error */
@@ -564,12 +565,13 @@ static int verity_verify_io(struct dm_verity_io *io)if (v->validated_blocks)set_bit (cur_block, v->validated_blocks);kunmap_local (data);continue ; } - r =verity_handle_data_hash_mismatch (v, io, bio, cur_block, - data); + r =verity_handle_data_hash_mismatch (v, io, bio, +verity_io_want_digest (v, io), + cur_block, data);kunmap_local (data);if (unlikely (r))return r; }
drivers/md/dm-verity.h +1 -0
@@ -93,10 +93,11 @@ struct dm_verity_io {bool in_bh;struct work_struct work;struct work_struct bh_work; +u8 tmp_digest[HASH_MAX_DIGESTSIZE ];u8 real_digest[HASH_MAX_DIGESTSIZE ];u8 want_digest[HASH_MAX_DIGESTSIZE ];/* * This struct is followed by a variable-sized hash request of size
Signature · 1 line
2.46.2
Eric Biggers
maintainerPATCH v7 7/7 dm-verity: improve performance by using multibuffer hashing
In reply to Eric BiggersFrom: Eric Biggers <[email protected]>
When supported by the hash algorithm, use crypto_shash_finup_mb() to interleave the hashing of pairs of data blocks. On some CPUs this nearly doubles hashing performance. The increase in overall throughput of cold-cache dm-verity reads that I'm seeing on arm64 and x86_64 is roughly 35% (though this metric is hard to measure as it jumps around a lot).
For now this is only done on data blocks, not Merkle tree blocks. We could use finup_mb on Merkle tree blocks too, but that is less important as there aren't as many Merkle tree blocks as data blocks, and that would require some additional code restructuring.
--- drivers/md/dm-verity-target.c | 166 ++++++++++++++++++++++++++-------- drivers/md/dm-verity.h | 33 ++++--- 2 files changed, 147 insertions(+), 52 deletions(-)
drivers/md/dm-verity-target.c +128 -38
@@ -182,22 +182,28 @@ static int verity_ahash_final(struct dm_verity *v, struct ahash_request *req, r =crypto_wait_req (crypto_ahash_final (req), wait); out:return r; } +static int verity_ahash (struct dm_verity *v,struct dm_verity_io *io, +const u8 *data,size_t len,u8 *digest,bool may_sleep) +{ +struct ahash_request *req =verity_io_hash_req (v, io); +struct crypto_wait wait; + +return verity_ahash_init (v, req, &wait, may_sleep) ?: +verity_ahash_update (v, req, data, len, &wait) ?: +verity_ahash_final (v, req, digest, &wait); +} +int verity_hash (struct dm_verity *v,struct dm_verity_io *io,const u8 *data,size_t len,u8 *digest,bool may_sleep) {int r;if (static_branch_unlikely (&ahash_enabled) && !v->shash_tfm) { -struct ahash_request *req =verity_io_hash_req (v, io); -struct crypto_wait wait; - - r =verity_ahash_init (v, req, &wait, may_sleep) ?: -verity_ahash_update (v, req, data, len, &wait) ?: -verity_ahash_final (v, req, digest, &wait); + r =verity_ahash (v, io, data, len, digest, may_sleep); }else {struct shash_desc *desc =verity_io_hash_req (v, io); desc->tfm = v->shash_tfm; r = crypto_shash_import(desc, v->initial_hashstate) ?:
@@ -206,10 +212,38 @@ int verity_hash(struct dm_verity *v, struct dm_verity_io *io,if (unlikely (r))DMERR ("Error hashing block: %d" , r);return r; } +static int verity_hash_mb (struct dm_verity *v,struct dm_verity_io *io, +const u8 *data[],size_t len,u8 *digests[], +int num_blocks) +{ +int r =0 ; + +if (static_branch_unlikely (&ahash_enabled) && !v->shash_tfm) { +int i; + +/* Note: in practice num_blocks is always 1 in this case. */ +for (i =0 ; i < num_blocks; i++) { + r =verity_ahash (v, io, data[i], len, digests[i], + !io->in_bh); +if (r) +break ; + } + }else { +struct shash_desc *desc =verity_io_hash_req (v, io); + + desc->tfm = v->shash_tfm; + r =crypto_shash_import (desc, v->initial_hashstate) ?: +crypto_shash_finup_mb (desc, data, len, digests, num_blocks); + } +if (unlikely (r)) +DMERR ("Error hashing blocks: %d" , r); +return r; +} +static void verity_hash_at_level (struct dm_verity *v,sector_t block,int level,sector_t *hash_block,unsigned int *offset) {sector_t position =verity_position_at_level (v, block, level);unsigned int idx;
@@ -457,13 +491,16 @@ static noinline int verity_recheck(struct dm_verity *v, struct dm_verity_io *io, }static int verity_handle_data_hash_mismatch (struct dm_verity *v,struct dm_verity_io *io,struct bio *bio, -const u8 *want_digest, -sector_t blkno,u8 *data) +struct pending_block *block) { +const u8 *want_digest = block->want_digest; +sector_t blkno = block->blkno; +u8 *data = block->data; +if (static_branch_unlikely (&use_bh_wq_enabled) && io->in_bh) {/* * Error handling code (FEC included) cannot be run in the * BH workqueue, so fallback to a standard workqueue. */
@@ -487,10 +524,57 @@ static int verity_handle_data_hash_mismatch(struct dm_verity *v,return -EIO ; }return 0 ; } +static void verity_clear_pending_blocks (struct dm_verity_io *io) +{ +int i; + +for (i = io->num_pending -1 ; i >=0 ; i--) { +kunmap_local (io->pending_blocks[i].data); + io->pending_blocks[i].data =NULL ; + } + io->num_pending =0 ; +} + +static int verity_verify_pending_blocks (struct dm_verity *v, +struct dm_verity_io *io, +struct bio *bio) +{ +const u8 *data[DM_VERITY_MAX_PENDING_DATA_BLOCKS ]; +u8 *real_digests[DM_VERITY_MAX_PENDING_DATA_BLOCKS ]; +int i; +int r; + +for (i =0 ; i < io->num_pending; i++) { + data[i] = io->pending_blocks[i].data; + real_digests[i] = io->pending_blocks[i].real_digest; + } + + r =verity_hash_mb (v, io, data,1 << v->data_dev_block_bits, + real_digests, io->num_pending); +if (unlikely (r)) +return r; + +for (i =0 ; i < io->num_pending; i++) { +struct pending_block *block = &io->pending_blocks[i]; + +if (likely (memcmp (block->real_digest, block->want_digest, + v->digest_size) ==0 )) { +if (v->validated_blocks) +set_bit (block->blkno, v->validated_blocks); + }else { + r =verity_handle_data_hash_mismatch (v, io, bio, block); +if (unlikely (r)) +return r; + } + } +verity_clear_pending_blocks (io); +return 0 ; +} +/* * Verify one "dm_verity_io" structure. */ static int verity_verify_io (struct dm_verity_io *io) {
@@ -498,10 +582,13 @@ static int verity_verify_io(struct dm_verity_io *io)const unsigned int block_size =1 << v->data_dev_block_bits;struct bvec_iter iter_copy;struct bvec_iter *iter;struct bio *bio =dm_bio_from_per_bio_data (io, v->ti->per_io_data_size);unsigned int b; +int r; + + io->num_pending =0 ;if (static_branch_unlikely (&use_bh_wq_enabled) && io->in_bh) {/* * Copy the iterator in case we need to restart * verification in a work-queue.
@@ -511,36 +598,38 @@ static int verity_verify_io(struct dm_verity_io *io) }else iter = &io->iter;for (b =0 ; b < io->n_blocks; b++,bio_advance_iter (bio, iter, block_size)) { -int r; -sector_t cur_block = io->block + b; +sector_t blkno = io->block + b; +struct pending_block *block;bool is_zero;struct bio_vec bv;void *data;if (v->validated_blocks && bio->bi_status ==BLK_STS_OK && -likely (test_bit (cur_block, v->validated_blocks))) +likely (test_bit (blkno, v->validated_blocks)))continue ; - r =verity_hash_for_block (v, io, cur_block, -verity_io_want_digest (v, io), + block = &io->pending_blocks[io->num_pending]; + + r =verity_hash_for_block (v, io, blkno, block->want_digest, &is_zero);if (unlikely (r <0 )) -return r; + goto error; bv =bio_iter_iovec (bio, *iter);if (unlikely (bv.bv_len < block_size)) {/* * Data block spans pages. This should not happen, * since dm-verity sets dma_alignment to the data block * size minus 1, and dm-verity also doesn't allow the * data block size to be greater than PAGE_SIZE. */ DMERR_LIMIT ("unaligned io (data block spans pages)" ); -return -EIO ; + r = -EIO ; + goto error; } data =bvec_kmap_local (&bv);if (is_zero) {
@@ -550,34 +639,30 @@ static int verity_verify_io(struct dm_verity_io *io)*/ memset (data,0 , block_size);kunmap_local (data);continue ; } - - r =verity_hash (v, io, data, block_size, -verity_io_real_digest (v, io), !io->in_bh); -if (unlikely (r <0 )) { -kunmap_local (data); -return r; + block->data = data; + block->blkno = blkno; +if (++io->num_pending == v->mb_max_msgs) { + r =verity_verify_pending_blocks (v, io, bio); +if (unlikely (r)) + goto error; } + } -if (likely (memcmp (verity_io_real_digest (v, io), -verity_io_want_digest (v, io), v->digest_size) ==0 )) { -if (v->validated_blocks) -set_bit (cur_block, v->validated_blocks); -kunmap_local (data); -continue ; - } - r =verity_handle_data_hash_mismatch (v, io, bio, -verity_io_want_digest (v, io), - cur_block, data); -kunmap_local (data); +if (io->num_pending) { + r =verity_verify_pending_blocks (v, io, bio);if (unlikely (r)) -return r; + goto error; }return 0 ; + +error: +verity_clear_pending_blocks (io); +return r; }/* * Skip verity work in response to I/O error when system is shutting down. */
@@ -1217,14 +1302,15 @@ static int verity_setup_hash_alg(struct dm_verity *v, const char *alg_name)/* * Allocate the hash transformation object that this dm-verity instance * will use. The vast majority of dm-verity users use CPU-based * hashing, so when possible use the shash API to minimize the crypto -* API overhead. If the ahash API resolves to a different driver -* (likely an off-CPU hardware offload), use ahash instead. Also use -* ahash if the obsolete dm-verity format with the appended salt is -* being used, so that quirk only needs to be handled in one place. +* API overhead, especially when multibuffer hashing is used. If the +* ahash API resolves to a different driver (likely an off-CPU hardware +* offload), use ahash instead. Also use ahash if the obsolete +* dm-verity format with the appended salt is being used, so that quirk +* only needs to be handled in one place. */ ahash =crypto_alloc_ahash (alg_name,0 , v->use_bh_wq ?CRYPTO_ALG_ASYNC :0 );if (IS_ERR (ahash)) { ti->error ="Cannot initialize hash function" ;
@@ -1248,17 +1334,21 @@ static int verity_setup_hash_alg(struct dm_verity *v, const char *alg_name) ahash =NULL ; v->shash_tfm = shash; v->digest_size =crypto_shash_digestsize (shash); v->hash_reqsize =sizeof (struct shash_desc ) +crypto_shash_descsize (shash); -DMINFO ("%s using shash \"%s\"" , alg_name, driver_name); + v->mb_max_msgs =min (crypto_shash_mb_max_msgs (shash), +DM_VERITY_MAX_PENDING_DATA_BLOCKS ); +DMINFO ("%s using shash \"%s\"%s" , alg_name, driver_name, + v->mb_max_msgs >1 ?" (multibuffer)" :"" ); }else { v->ahash_tfm = ahash;static_branch_inc (&ahash_enabled); v->digest_size =crypto_ahash_digestsize (ahash); v->hash_reqsize =sizeof (struct ahash_request ) +crypto_ahash_reqsize (ahash); + v->mb_max_msgs =1 ;DMINFO ("%s using ahash \"%s\"" , alg_name, driver_name); }if ((1 << v->hash_dev_block_bits) < v->digest_size *2 ) { ti->error ="Digest size too big" ;return -EINVAL ;
drivers/md/dm-verity.h +19 -14
@@ -59,10 +59,11 @@ struct dm_verity {unsigned char hash_per_block_bits;/* log2(hashes in hash block) */ unsigned char levels;/* the number of tree levels */ unsigned char version; bool hash_failed:1; /* set if hash of any block failed */ bool use_bh_wq:1; /* try to verify in BH wq before normal work-queue */ +unsigned char mb_max_msgs;/* max multibuffer hashing interleaving factor */ unsigned int digest_size;/* digest size for the current hash algorithm */ unsigned int hash_reqsize;/* the size of temporary space for crypto */ enum verity_mode mode;/* mode for handling verification errors */ unsigned int corrupted_errs;/* Number of errors for corrupted blocks */
@@ -78,10 +79,19 @@ struct dm_verity {struct dm_io_client *io;mempool_t recheck_pool; }; +#define DM_VERITY_MAX_PENDING_DATA_BLOCKS HASH_MAX_MB_MSGS + +struct pending_block { +void *data; +sector_t blkno; +u8 want_digest[HASH_MAX_DIGESTSIZE ]; +u8 real_digest[HASH_MAX_DIGESTSIZE ]; +}; +struct dm_verity_io {struct dm_verity *v;/* original value of bio->bi_end_io */ bio_end_io_t *orig_bi_end_io;
@@ -94,12 +104,19 @@ struct dm_verity_io {struct work_struct work;struct work_struct bh_work;u8 tmp_digest[HASH_MAX_DIGESTSIZE ]; -u8 real_digest[HASH_MAX_DIGESTSIZE ]; -u8 want_digest[HASH_MAX_DIGESTSIZE ]; + +/* +* This is the queue of data blocks that are pending verification. We +* allow multiple blocks to be queued up in order to support multibuffer +* hashing, i.e. interleaving the hashing of multiple messages. On many +* CPUs this improves performance significantly. +*/ +int num_pending; +struct pending_block pending_blocks[DM_VERITY_MAX_PENDING_DATA_BLOCKS ];/* * This struct is followed by a variable-sized hash request of size * v->hash_reqsize, either a struct ahash_request or a struct shash_desc * (depending on whether ahash_tfm or shash_tfm is being used). To
@@ -111,22 +128,10 @@ static inline void *verity_io_hash_req(struct dm_verity *v, struct dm_verity_io *io) {return io +1 ; } -static inline u8 *verity_io_real_digest (struct dm_verity *v, -struct dm_verity_io *io) -{ -return io->real_digest; -} - -static inline u8 *verity_io_want_digest (struct dm_verity *v, -struct dm_verity_io *io) -{ -return io->want_digest; -} -extern int verity_hash (struct dm_verity *v,struct dm_verity_io *io,const u8 *data,size_t len,u8 *digest,bool may_sleep);extern int verity_hash_for_block (struct dm_verity *v,struct dm_verity_io *io,sector_t block,u8 *digest,bool *is_zero);
Signature · 1 line
2.46.2